/***************************************************************************
*  T B f f o l d e r                                                       *
*                                                                          *
*  Author..: Torsten Johann                                                *
*  E-mail..: t-johann@gmx.de                                               *
*  Date....: 13. Jan. 2000                                                 *
*  Version.: 1.00                                                          *
*  Status..: FREEWARE                                                      *
*  Function: Implementation of Win32 API function 'SHBrowseForFolder'      *
*            in Borland C++Builder 3.0                                     *
*                                                                          *
*  (C) Copyright 2000 Torsten Johann                                       *
*                                                                          *
***************************************************************************/

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "Bffolder.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------
// ValidCtrCheck wird verwendet, um sicherzustellen, da die erzeugten Komponenten
// keine rein virtuellen Funktionen besitzen.
//

static inline void ValidCtrCheck(TBffolder *)
{
  new TBffolder(NULL);
}
//---------------------------------------------------------------------------
__fastcall TBffolder::TBffolder(TComponent* Owner)
    : TCommonDialog(Owner)
{
  GetDllVersion ();
  FRootDir  = rdDesktop;
  FEnableOK = true;
  FOptions  << bffStatusText;
}
//---------------------------------------------------------------------------
void __fastcall TBffolder::GetDllVersion (void)
{
  HINSTANCE hinstDll;

  FDllVersion = 0;
  hinstDll = LoadLibrary ("Shell32.dll");
  if (hinstDll)
  {
    DLLGETVERSIONPROC pDllGetVersion;
    pDllGetVersion = (DLLGETVERSIONPROC)GetProcAddress (hinstDll, "DllGetVersion");
    /*Because some DLLs may not implement this function, you
     *must test for it explicitly. Depending on the particular
     *DLL, the lack of a DllGetVersion function may
     *be a useful indicator of the version.
     */
    if (pDllGetVersion)
    {
      DLLVERSIONINFO dvi;
      HRESULT        hr;

      ZeroMemory (&dvi, sizeof(dvi));
      dvi.cbSize = sizeof(dvi);
      hr = (*pDllGetVersion)(&dvi);

      if (SUCCEEDED(hr))
      {
         FDllVersion = PACKVERSION(dvi.dwMajorVersion, dvi.dwMinorVersion);
      }
    }
    FreeLibrary(hinstDll);
  }
}
//---------------------------------------------------------------------------
int __stdcall BffolderCallbackProc (HWND hwnd, UINT uMsg, LPARAM lParam, LPARAM lpData)
{
  TBffolder* bff;
  char       szPath [MAX_PATH];
  AnsiString asNewSel;
  AnsiString asInvalidName;
  bool       Continue = false;
  STRRET     lpName;

  if (uMsg == BFFM_INITIALIZED || uMsg == BFFM_SELCHANGED ||
      uMsg == BFFM_VALIDATEFAILED)
  {
    bff = (TBffolder*)lpData;
    if (bff == NULL)
      return 0;

    bff->FhWnd = hwnd;
    bff->FMsg = uMsg;
    bff->FlParam = lParam;
    bff->FlpData = lpData;

    if (uMsg == BFFM_INITIALIZED)
    {
      // change caption, if spezified
      if (!bff->Caption.IsEmpty ())
        SetWindowText (hwnd, bff->Caption.c_str ());

      // set status text
      SendMessage (hwnd, BFFM_SETSTATUSTEXT, 0,
                   (LPARAM)bff->StatusText.c_str ());

      // Hilight <SelectedDir>
      if (!bff->SelectedDir.IsEmpty ())
        SendMessage (bff->FhWnd, BFFM_SETSELECTION, true,
                     (LPARAM)bff->SelectedDir.c_str ());

      // set OK button state
      SendMessage (hwnd, BFFM_ENABLEOK, 0, bff->EnableOK);

      // Invoke event handler
      if (bff->OnInitialized)
        bff->OnInitialized (bff);
    }

    if (uMsg == BFFM_SELCHANGED)
    {
      if (bff->OnSelectionChanged)
      {
        // get new display name
        if (bff->lpShellFolder->GetDisplayNameOf (LPCITEMIDLIST(lParam),
            SHGDN_NORMAL, &lpName) == NOERROR)
        {
          switch (lpName.uType)
          {
            case STRRET_CSTR:
              bff->DisplayName = AnsiString (lpName.cStr);
              break;
            case STRRET_OFFSET:
              bff->DisplayName = AnsiString ((char*)lParam + lpName.uOffset);
              break;
            case STRRET_WSTR:
              bff->DisplayName = AnsiString ((char*)lpName.pOleStr);
            default:
              break;
          }
        }
        // get new selection
        if (SHGetPathFromIDList (LPCITEMIDLIST(lParam), szPath))
        {
          asNewSel = AnsiString (szPath);
        }
        // invoke event handler
        bff->OnSelectionChanged (bff, asNewSel);
      }
    }

    if (uMsg == BFFM_VALIDATEFAILED)
    {
      if (bff->FDllVersion >= PACKVERSION(4,71))
      {
        if (bff->OnValidateFailed)
        {
          asInvalidName = AnsiString ((char*) lParam);
          bff->OnValidateFailed (bff, asInvalidName, Continue);
        }
      }
      return Continue;
    }
  }
  return 0;
}
//---------------------------------------------------------------------------
bool __fastcall TBffolder::Execute (void)
{
  char         szFolderName [MAX_PATH];
  int          nRootDir;
  LPITEMIDLIST dummy = NULL;
  ULONG        uEaten;
  WCHAR        wsUserRootDir [MAX_PATH];

  if (SHGetMalloc (&lpMalloc) != NOERROR)
    return false;
  if (SHGetDesktopFolder (&lpShellFolder) != NOERROR)
    return false;

  ZeroMemory (&bi, sizeof (BROWSEINFO));

  if (FOptions.Contains (bffBrowseForComputer))
    bi.ulFlags += BIF_BROWSEFORCOMPUTER;
  if (FOptions.Contains (bffBrowseForPrinter))
    bi.ulFlags += BIF_BROWSEFORPRINTER;
  if (FOptions.Contains (bffDontGoBelowDomain))
    bi.ulFlags += BIF_DONTGOBELOWDOMAIN;
  if (FOptions.Contains (bffReturnFSAncestors))
    bi.ulFlags += BIF_RETURNFSANCESTORS;
  if (FOptions.Contains (bffReturnOnlyFSDirs))
    bi.ulFlags += BIF_RETURNONLYFSDIRS;
  if (FOptions.Contains (bffStatusText))
    bi.ulFlags += BIF_STATUSTEXT;

  if (FDllVersion >= PACKVERSION(4,71))
  {
    if (FOptions.Contains (bffValidate))
      bi.ulFlags += BIF_VALIDATE;
    if (FOptions.Contains (bffEditBox))
      bi.ulFlags += BIF_EDITBOX;
    if (FOptions.Contains (bffBrowseIncludeFiles))
      bi.ulFlags += BIF_BROWSEINCLUDEFILES;
  }

  switch (FRootDir)
  {
    case rdRecycleBin:             nRootDir = CSIDL_BITBUCKET; break;
    case rdControlPanel:           nRootDir = CSIDL_CONTROLS; break;
    case rdDesktop:                nRootDir = CSIDL_DESKTOP; break;
    case rdDesktopDirectory:       nRootDir = CSIDL_DESKTOPDIRECTORY; break;
    case rdMyComputer:             nRootDir = CSIDL_DRIVES; break;
    case rdFonts:                  nRootDir = CSIDL_FONTS; break;
    case rdNetHood:                nRootDir = CSIDL_NETHOOD; break;
    case rdNetwork:                nRootDir = CSIDL_NETWORK; break;
    case rdMyDocuments:            nRootDir = CSIDL_PERSONAL; break;
    case rdPrinters:               nRootDir = CSIDL_PRINTERS; break;
    case rdPrograms:               nRootDir = CSIDL_PROGRAMS; break;
    case rdRecent:                 nRootDir = CSIDL_RECENT; break;
    case rdSendTo:                 nRootDir = CSIDL_SENDTO; break;
    case rdStartMenu:              nRootDir = CSIDL_STARTMENU; break;
    case rdStartup:                nRootDir = CSIDL_STARTUP; break;
    case rdTemplates:              nRootDir = CSIDL_TEMPLATES; break;
    case rdInternet:               nRootDir = CSIDL_INTERNET; break;
    case rdFavorites:              nRootDir = CSIDL_FAVORITES; break;
    case rdCommonStartMenu:        nRootDir = CSIDL_COMMON_STARTMENU; break;
    case rdCommonPrograms:         nRootDir = CSIDL_COMMON_PROGRAMS; break;
    case rdCommonStartup:          nRootDir = CSIDL_COMMON_STARTUP; break;
    case rdCommonDesktopDirectory: nRootDir = CSIDL_COMMON_DESKTOPDIRECTORY; break;
    case rdAppData:                nRootDir = CSIDL_APPDATA; break;
    case rdPrintHood:              nRootDir = CSIDL_PRINTHOOD; break;
    case rdAltStartup:             nRootDir = CSIDL_ALTSTARTUP; break;
    case rdCommonAltStartup:       nRootDir = CSIDL_COMMON_ALTSTARTUP; break;
    case rdCommonFavorites:        nRootDir = CSIDL_COMMON_FAVORITES; break;
    case rdInternetCache:          nRootDir = CSIDL_INTERNET_CACHE; break;
    case rdCookies:                nRootDir = CSIDL_COOKIES; break;
    case rdHistory:                nRootDir = CSIDL_HISTORY; break;
    case rdUseSelectedDir:
      if (!SelectedDir.IsEmpty ())
      {
        MultiByteToWideChar (CP_ACP, MB_PRECOMPOSED, FSelectedDir.c_str(), -1, wsUserRootDir, MAX_PATH);
        if (lpShellFolder->ParseDisplayName (0, NULL, wsUserRootDir,
              &uEaten, &dummy, NULL) != NOERROR)
        {
          dummy = NULL;
          nRootDir = CSIDL_DESKTOP;
        }
        else
          bi.pidlRoot = dummy;
      }
      break;
  }
  if (nRootDir != CSIDL_DESKTOP && !dummy)
  {
    if (SHGetSpecialFolderLocation (Application->Handle, nRootDir, &dummy) == NOERROR)
    {
      bi.pidlRoot = dummy;
    }
  }

  bi.hwndOwner = Application->Handle;
  bi.pszDisplayName = szFolderName;
  bi.lpszTitle = FTitle.c_str ();
  bi.lpfn = BffolderCallbackProc;
  bi.lParam = (long)this;

  ItemID = SHBrowseForFolder (&bi);
  if (dummy)
  {
    lpMalloc->Free (dummy);
    lpMalloc->Release ();
  }
  lpShellFolder->Release ();
  if (ItemID != NULL)
  {
    FDisplayName = AnsiString (szFolderName);
    SHGetPathFromIDList (ItemID, szFolderName);
    FSelectedDir = AnsiString (szFolderName);
    lpMalloc->Free (ItemID);
    lpMalloc->Release ();
    return true;
  }
  return false;
}
//---------------------------------------------------------------------------
void __fastcall TBffolder::SetSelectedDir (const AnsiString AName)
{
  FSelectedDir = AName;
  if (FhWnd != NULL)
    SendMessage (FhWnd, BFFM_SETSELECTION, true, (LPARAM)AName.c_str ());
}
//---------------------------------------------------------------------------
void __fastcall TBffolder::SetCaption (const AnsiString ACaption)
{
  if (!ACaption.IsEmpty ())
  {
    FCaption = ACaption;
    if (FhWnd != NULL)
      SetWindowText (FhWnd, FCaption.c_str ());
  }
}
//---------------------------------------------------------------------------
AnsiString _fastcall TBffolder::GetCaption (void)
{
  char szWindowText [MAX_PATH];

  if ( (FhWnd != NULL) && (FCaption.IsEmpty ()) )
  {
    GetWindowText (FhWnd, szWindowText, MAX_PATH);
    FCaption = AnsiString (szWindowText);
  }
  return FCaption;
}
//---------------------------------------------------------------------------
void __fastcall TBffolder::SetStatusText (const AnsiString AStatusText)
{
  FStatusText = AStatusText;
  if (FhWnd != NULL)
    SendMessage (FhWnd, BFFM_SETSTATUSTEXT, 0, (LPARAM)FStatusText.c_str ());
}
//---------------------------------------------------------------------------
void __fastcall TBffolder::SetEnableOK (bool AEnable)
{
  FEnableOK = AEnable;
  if (FhWnd != NULL)
    SendMessage (FhWnd, BFFM_ENABLEOK, 0, AEnable);
}
//---------------------------------------------------------------------------
namespace Bffolder
{
  void __fastcall PACKAGE Register()
  {
    TComponentClass classes[1] = {__classid(TBffolder)};
    RegisterComponents("Win32", classes, 0);
  }
}

